#pragma once

#include <map>
#include <string>
#include <WinSock2.h>
#include <thread>
#include <mutex>
#include <list>
#include <utility>

template<typename KeyType, typename ValueType>
class SharedMap
{
public:
	SharedMap();
	~SharedMap();
	bool findKey(const KeyType& key, ValueType& returnValue);
	void setValue(const KeyType& key, const ValueType& newValue);
	list<pair<KeyType, ValueType> > getAllEntries() const;

private:
	// Disallow the copy constructor (we can't copy the lock without reference counting)
	SharedMap(const SharedMap& orig);

	std::map<KeyType, ValueType> map_;

	std::mutex* lock_;
};

template<typename KeyType, typename ValueType>
SharedMap<KeyType, ValueType>::SharedMap() : lock_(new mutex)
{
}


template<typename KeyType, typename ValueType>
SharedMap<KeyType, ValueType>::~SharedMap()
{
	delete lock_;
}

template<typename KeyType, typename ValueType>
bool SharedMap<KeyType, ValueType>::findKey(const KeyType& key, 
	ValueType& lookupValue) 
{
	lock_->lock();
	bool hasTheKey = (map_.find(key) != map_.end());
	if (hasTheKey)
	{
		lookupValue = map_[key];
	}
	lock_->unlock();

	return hasTheKey;
}

template<typename KeyType, typename ValueType>
void SharedMap<KeyType, ValueType>::setValue(const KeyType& key,
	const ValueType& newValue)
{
	lock_->lock();
	map_[key] = newValue;
	lock_->unlock();
}

template<typename KeyType, typename ValueType>
list<pair<KeyType, ValueType> > SharedMap<KeyType, ValueType>::getAllEntries() const
{
	list<pair<KeyType, ValueType> > returnList;

	lock_->lock();
	for (map<KeyType, ValueType>::const_iterator it = map_.begin(); it != map_.end(); ++it)
	{ 
		returnList.push_front(make_pair(it->first, it->second));
	}
	lock_->unlock();
	return returnList;
}