//#define _AFXDLL
#define _WINSOCK_DEPRECATED_NO_WARNINGS
//#include <afxwin.h>
#include <winsock2.h>
#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>
using namespace std;

#include "NameQuery.h"
#include "NameResolution.h"
#include <Ws2tcpip.h>
#include "SharedMap.h"

string twoDigitString(int number)
{
	stringstream stream;
	stream << setfill('0') << setw(3) << number;
	return stream.str();
}

list<pair<string, in_addr> > getAll113Hosts()
{
	list<pair<string, in_addr> > hostList;
	const int n113Hosts = 25;
	for (int i = 0; i < n113Hosts; i++)
	{
		string domainName = "l-hobb-0113-" + twoDigitString(i);
		hostent* hostEntry = gethostbyname(domainName.c_str());
		if (hostEntry != 0)
		{
			hostList.push_front(make_pair(domainName,
				*reinterpret_cast<in_addr*>(hostEntry->h_addr_list[0])));
		}
	}
	return hostList;
}

void serverThread(SharedMap<string, in_addr>* map)
{
	SOCKET serverUDP = socket(AF_INET, SOCK_DGRAM, 0);

	int replyLength = sizeof(sockaddr);
	NameQuery request;
	sockaddr_in bindAddress;
	bindAddress.sin_family = AF_INET;
	bindAddress.sin_port = htons(51002);
	bindAddress.sin_addr.s_addr = htonl(INADDR_ANY);

	int bindStatus = ::bind(serverUDP, reinterpret_cast<sockaddr*>(&bindAddress),
		sizeof(bindAddress));

	sockaddr_in replyAddress;

	NameResolution response;
	while (true)
	{
		// Receive a request for resolution
		int nBytesReceived = recvfrom(serverUDP,
			reinterpret_cast<char*>(&request),
			sizeof(request), 0,
			reinterpret_cast<sockaddr*>(&replyAddress),
			&replyLength);

		cout << "request contained " << nBytesReceived << " bytes." << endl;

		const int MAX_IP_LENGTH = 128;
		char hostIP[MAX_IP_LENGTH + 1];

		// Write the IP address of the requester
		// inet_ntoa is deprecated.
		//cerr << "Received request from " << inet_ntoa(replyAddress.sin_addr) 
		//	 << " for " << request.getNameToResolve() << endl;
		inet_ntop(AF_INET, reinterpret_cast<PVOID>(&replyAddress.sin_addr), 
			hostIP, MAX_IP_LENGTH);
		cerr << "Received request from " << hostIP
			<< " for " << request.getNameToResolve() << endl;

		response.setRequestNumber(request.getRequestNumber());
		in_addr ipAddress;
		if (map->findKey(request.getNameToResolve(), ipAddress))
		{
			response.setResolutionSuccessful(true);
			response.setIpAddressNBO(ipAddress);
			response.setName(request.getNameToResolve());
		}
		else
		{
			response.setResolutionSuccessful(false);
		}

		SOCKET sendSocket = socket(AF_INET, SOCK_DGRAM, 0);

		sendto(serverUDP,
			reinterpret_cast<const char*>(&response), sizeof(response), 0,
			reinterpret_cast<const sockaddr*>(&replyAddress), sizeof(replyAddress));
	}
}

int main(int argc, char**argv)
{
	getAll113Hosts();
	// udpClient someDomain.com 51029
	const int DOMAIN_NAME_ARG = 1;
	const int PORT_ARG = 2;
	const int NAME_ARG = 3;
	const int N_ARGS = NAME_ARG;

	if (argc != N_ARGS + 1)
	{
		cerr << "usage: " << argv[0] << " <ipAddressOfPeer> <portNumber> <name>" << endl;
	}
	else
	{

		WORD versionRequired = 0x0101;
		WSADATA wsadata;

		int startupStatus = WSAStartup(versionRequired,
			&wsadata);



		if (startupStatus == 0)
		{
			SharedMap<string, in_addr> map;

			// Map contains ip addresses for domain names
			// IN_ADDR is really just an int

			const int N_NAMES_TO_RESOLVE = 14;
			string namesToResolve[N_NAMES_TO_RESOLVE] =
			{ "JeffA", "AndrewB", "WillF",
			"RyanF", "ChandlerH", "RyanJ",
			"JonL", "JeffM", "JoshR",
			"EmmanuelS", "TravisS", "JoelW",
			"ThomasW", "RandyR" };


			// Get the name of this host
			const int MAX_HOSTNAME_LENGTH = 256;
			char hostname[MAX_HOSTNAME_LENGTH];

			// Get the IP address of this host
			gethostname(hostname, MAX_HOSTNAME_LENGTH);
			hostent* thisMachine = gethostbyname(hostname);

			// Enter the IP address in this table
			map.setValue(argv[NAME_ARG],
				*reinterpret_cast<IN_ADDR*>(thisMachine->h_addr_list[0]));

			std::thread serverThreadObject(serverThread, &map);

			// socket call specifies UDP with SOCK_DGRAM
			// or TCP with SOCK_STREAM
			// SOCK_RAW for just data
			SOCKET firstUDP = socket(AF_INET, SOCK_DGRAM, 0);

			sockaddr_in serverAddress;

			hostent* server = gethostbyname(
				argv[DOMAIN_NAME_ARG]);

			serverAddress.sin_addr =
				*reinterpret_cast<IN_ADDR*>(
				server->h_addr_list[0]);

			short portNumber = atoi(argv[PORT_ARG]);
			short portNumberInNBO = htons(portNumber);
			serverAddress.sin_port = portNumberInNBO;

			serverAddress.sin_family = AF_INET;

			int requestNumber = 1;

			NameQuery request(htonl(requestNumber), "RandyR");

			int nIterations = 10;
			int j = 0;
			while (j < nIterations)
			{
				// Create a list of ipAddresses to check
				list<pair<string, in_addr> > entries = map.getAllEntries();

				for (list<pair<string, in_addr> >::iterator it = entries.begin(); 
					it != entries.end();
					++it)
				{

					serverAddress.sin_addr =
						*reinterpret_cast<IN_ADDR*>(&it->second);

				for (int i = 0; i < N_NAMES_TO_RESOLVE; ++i)
				{
					request.setNameToResolve(namesToResolve[i]);
					request.setRequestNumber(requestNumber++);

					int nBytesAccept = sendto(firstUDP,
						reinterpret_cast<char*>(&request),
						sizeof(request), 0,
						reinterpret_cast<sockaddr*>(&serverAddress),
						sizeof(serverAddress));

					int replyLength = sizeof(sockaddr);
					NameResolution response;
					DWORD millisecondTimeout = 500;
					setsockopt(firstUDP, SOL_SOCKET, SO_RCVTIMEO,
						reinterpret_cast<const char*>(&millisecondTimeout), sizeof(DWORD));

					sockaddr_in replyAddress;
					int nBytesReceived = recvfrom(firstUDP,
						reinterpret_cast<char*>(&response),
						sizeof(NameResolution), 0,
						reinterpret_cast<sockaddr*>(&replyAddress),
						&replyLength);

					if (nBytesReceived > 0)
					{
						cout << "Got " << nBytesReceived << " bytes." << endl;
						cout << "Request Number: " << response.getRequestNumber() << endl;
						if (response.resolutionSuccessful())
						{
							cout << "Success: true" << endl;
							cout << "Name: " << response.getName();
							const int MAX_IP_LENGTH = 128;
							char hostIP[MAX_IP_LENGTH + 1];
							inet_ntop(AF_INET, reinterpret_cast<PVOID>(&replyAddress.sin_addr),
								hostIP, MAX_IP_LENGTH);
							cout << " ipAddress is " << hostIP << endl;
							map.setValue(response.getName(),
								response.getIpAddressInNBO());
						}
						else
						{
							cout << "Success: false" << endl;
						}
					}
					else
					{
						cout << "Recv status is " << nBytesReceived << " bytes." << endl;
					}
				}
				}
				++j;

				Sleep(2000);
			}

		serverThreadObject.join();
		}
		else
		{
			cerr << "Winsock startup error!" << endl;
		}
	}
	system("pause");
}